/// @copyright
/// Copyright (C) 2020 Assured Information Security, Inc.
///
/// @copyright
/// Permission is hereby granted, free of charge, to any person obtaining a copy
/// of this software and associated documentation files (the "Software"), to deal
/// in the Software without restriction, including without limitation the rights
/// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
/// copies of the Software, and to permit persons to whom the Software is
/// furnished to do so, subject to the following conditions:
///
/// @copyright
/// The above copyright notice and this permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// @copyright
/// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
/// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
/// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
/// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
/// SOFTWARE.

#ifndef BSL_DETAILS_FMT_IMPL_CSTR_TYPE_HPP
#define BSL_DETAILS_FMT_IMPL_CSTR_TYPE_HPP

#include "../cstr_type.hpp"
#include "../cstring.hpp"
#include "../fmt_options.hpp"
#include "../is_constant_evaluated.hpp"
#include "../safe_integral.hpp"
#include "../unlikely.hpp"
#include "fmt_impl_align.hpp"
#include "out.hpp"

namespace bsl
{
    /// <!-- description -->
    ///   @brief This function is responsible for implementing bsl::fmt
    ///     for cstr_type types. For strings, the only fmt options
    ///     that are available are alignment, fill and width, all of which
    ///     are handled by the fmt_impl_align_xxx functions.
    ///
    /// <!-- notes -->
    ///   @note This function exists in the details folder because it is
    ///     private to the BSL, but it does not exist in the details namespace
    ///     as it can be overridden by the user to provide their own
    ///     fmt support for their own types.
    ///
    /// <!-- inputs/outputs -->
    ///   @tparam T the type of out (i.e., debug, alert, etc)
    ///   @param o the instance of out<T> to output to
    ///   @param ops ops the fmt options used to format the output
    ///   @param str the cstr_type being outputted
    ///
    template<typename T>
    constexpr void
    fmt_impl(out<T> const o, fmt_options const &ops, cstr_type const str) noexcept
    {
        if (is_constant_evaluated()) {
            return;
        }

        if (unlikely(nullptr == str)) {
            constexpr bsl::cstr_type msg{"[empty bsl::cstr_type]"};
            details::fmt_impl_align_pre(o, ops, {}, true);
            o.write_to_console(msg, builtin_strlen(msg).get());
            details::fmt_impl_align_suf(o, ops, {}, true);
            return;
        }

        auto const len{bsl::builtin_strlen(str)};
        details::fmt_impl_align_pre(o, ops, len, true);
        o.write_to_console(str, builtin_strlen(str).get());
        details::fmt_impl_align_suf(o, ops, len, true);
    }

    /// <!-- description -->
    ///   @brief Outputs the provided cstr_type to the provided
    ///     output type.
    ///
    /// <!-- inputs/outputs -->
    ///   @tparam T the type of outputter provided
    ///   @param o the instance of the outputter used to output the value.
    ///   @param str the cstr_type to output
    ///   @return return o
    ///
    template<typename T>
    [[maybe_unused]] constexpr auto
    operator<<(out<T> const o, cstr_type const str) noexcept -> out<T>
    {
        if (is_constant_evaluated()) {
            return o;
        }

        if constexpr (o.empty()) {
            return o;
        }

        if (unlikely(nullptr == str)) {
            constexpr bsl::cstr_type msg{"[empty bsl::cstr_type]"};
            o.write_to_console(msg, builtin_strlen(msg).get());
            return o;
        }

        o.write_to_console(str, builtin_strlen(str).get());
        return o;
    }
}

#endif
